package ru.dbotthepony.mc.otm.graph.storage

import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.storage.IStorage
import ru.dbotthepony.mc.otm.storage.IStorageStack
import ru.dbotthepony.mc.otm.storage.StorageStackType
import java.util.*

open class BasicStorageGraphNode : IStorageGraphNode {
	private var resolver = LazyOptional.of<IStorageGraphNode> { this }
	private var valid = true
	protected val components = ArrayList<IStorage<*>>()

	override val storageNode = Graph6Node<IStorageGraphNode>(this)

	override fun fetchComponents(): List<IStorage<*>> {
		return Collections.unmodifiableList(components)
	}

	@Suppress("unchecked_cast")
	fun <T : IStorageStack, U : IStorage<T>> computeIfAbsent(identity: StorageStackType<T>, provider: (StorageStackType<T>) -> U): U {
		for (component in components) {
			if (component.storageType === identity) {
				return component as U
			}
		}

		val factory = provider(identity)
		addStorageComponent(factory)
		return factory
	}

	fun addStorageComponent(component: IStorage<*>) {
		for (component1 in components) {
			if (component === component1 || component1.storageType === component.storageType) {
				return
			}
		}

		components.add(component)
		storageGraph?.add(component)
	}

	fun invalidate() {
		if (!valid) return
		valid = false
		resolver.invalidate()
		// detachStorage();
	}

	fun revive() {
		if (valid) return
		valid = true
		resolver = LazyOptional.of { this }
		// attachStorage();
	}

	fun get(): LazyOptional<IStorageGraphNode> {
		return if (valid) resolver else LazyOptional.empty()
	}
}