package ru.dbotthepony.mc.otm.graph.storage

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.graph.Abstract6Graph
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.orNull
import ru.dbotthepony.mc.otm.storage.*

class StorageNetworkGraph : Abstract6Graph<IStorageGraphNode>() {
	private val virtualComponents = Object2ObjectArrayMap<StorageStackType<*>, VirtualComponent<*>>()

	fun <T : IStorageStack> insertStack(obj: T, simulate: Boolean): T {
		return (getVirtualComponent(StorageRegistry.get(obj::class.java)) as VirtualComponent<T>).insertStack(obj, simulate)
	}

	/**
	 * Returns a [VirtualComponent] representing [type] storage
	 */
	fun <T : IStorageStack> getVirtualComponent(type: StorageStackType<T>): VirtualComponent<T> {
		return virtualComponents.computeIfAbsent(type, Object2ObjectFunction { VirtualComponent(type) }) as VirtualComponent<T>
	}

	/**
	 * Returns a [VirtualComponent] representing [type] storage
	 */
	fun <T : IStorageStack> getVirtualComponent(type: Class<T>): VirtualComponent<T> {
		return virtualComponents.computeIfAbsent(StorageRegistry.get(type), Object2ObjectFunction { VirtualComponent(type) }) as VirtualComponent<T>
	}

	fun <T : IStorageStack> add(storage: IStorage<T>) {
		getVirtualComponent(storage.storageType).add(storage)
	}

	fun <T : IStorageStack> remove(storage: IStorage<T>) {
		getVirtualComponent(storage.storageType).remove(storage)
	}

	override fun onNodeAdded(node: Graph6Node<IStorageGraphNode>) {
		for (identity in node.value.fetchComponents()) {
			add(identity)
		}
	}

	override fun onNodeRemoved(node: Graph6Node<IStorageGraphNode>) {
		for (identity in node.value.fetchComponents()) {
			remove(identity)
		}
	}

	companion object {
		@JvmStatic
		fun discoverFull(tile: BlockEntity, node: Graph6Node<IStorageGraphNode>) {
			if (tile.level !is ServerLevel)
				return

			return discoverFull(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IStorageGraphNode>? {
					return _tile.getCapability(MatteryCapability.STORAGE_NODE).orNull()?.storageNode
				},
				::StorageNetworkGraph
			)
		}

		@JvmStatic
		fun discover(tile: BlockEntity, node: Graph6Node<IStorageGraphNode>): Boolean {
			if (tile.level !is ServerLevel)
				return false

			return discover(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IStorageGraphNode>? {
					return _tile.getCapability(MatteryCapability.STORAGE_NODE).orNull()?.storageNode
				},
				::StorageNetworkGraph
			)
		}
	}
}